import 'dart:convert';

class Protocol {
  final List<Domain> domains;

  Protocol.fromJson(Map json)
      : domains =
            (json['domains'] as List).map((j) => Domain.fromJson(j)).toList();

  factory Protocol.fromString(String protocol) =>
      Protocol.fromJson(jsonDecode(protocol));
}

class Domain {
  final String name;
  final String description;
  final List<ComplexType> types;
  final List<Command> commands;
  final List<Event> events;
  final bool deprecated;

  Domain.fromJson(Map json)
      : name = json['domain'],
        description = json['description'],
        types = json.containsKey('types')
            ? (json['types'] as List)
                .map((j) => ComplexType.fromJson(j))
                .toList()
            : const [],
        commands = json.containsKey('commands')
            ? (json['commands'] as List)
                .map((j) => Command.fromJson(j))
                .toList()
            : const [],
        events = json.containsKey('events')
            ? (json['events'] as List).map((j) => Event.fromJson(j)).toList()
            : const [],
        deprecated = json['deprecated'] ?? false;
}

class ComplexType {
  final String id;
  final String description;
  final String type;
  final List<Parameter> properties;
  final List<String> enums;
  final ListItems items;

  ComplexType(
      {this.id,
      this.properties = const [],
      this.description,
      this.type,
      this.enums,
      this.items});

  ComplexType.fromJson(Map json)
      : id = json['id'],
        description = json['description'],
        type = json['type'],
        properties = json.containsKey('properties')
            ? (json['properties'] as List)
                .map((j) => Parameter.fromJson(j))
                .toList()
            : const [],
        enums = (json['enum'] as List)?.cast<String>(),
        items = json.containsKey('items')
            ? ListItems.fromJson(json['items'])
            : null;
}

class Command {
  final String name;
  final String description;
  final List<Parameter> parameters;
  final List<Parameter> returns;
  final bool deprecated;

  Command.fromJson(Map json)
      : name = json['name'],
        description = json['description'],
        deprecated = json['deprecated'] ?? false,
        parameters = json.containsKey('parameters')
            ? (json['parameters'] as List)
                .map((j) => Parameter.fromJson(j))
                .toList()
            : const [],
        returns = json.containsKey('returns')
            ? (json['returns'] as List)
                .map((j) => Parameter.fromJson(j))
                .toList()
            : const [];
}

class Event {
  final String name;
  final String description;
  final List<Parameter> parameters;

  Event.fromJson(Map json)
      : name = json['name'],
        description = json['description'],
        parameters = json.containsKey('parameters')
            ? (json['parameters'] as List)
                .map((j) => Parameter.fromJson(j))
                .toList()
            : const [];
}

class Parameter implements Typed {
  final String name;
  final String description;
  final bool optional;
  final bool deprecated;
  final ListItems items;
  final List<String> enumValues;

  @override
  final String type;

  @override
  final String ref;

  Parameter(
      {this.name,
      this.description,
      this.type,
      this.ref,
      this.optional = false,
      this.deprecated = false,
      this.items,
      this.enumValues});

  Parameter.fromJson(Map json)
      : name = json['name'],
        description = json['description'],
        type = json['type'],
        ref = json[r'$ref'],
        optional = json['optional'] ?? false,
        deprecated = json['deprecated'] ?? false,
        items = json.containsKey('items')
            ? ListItems.fromJson(json['items'])
            : null,
        enumValues = json.containsKey('enum')
            ? (json['enum'] as List).cast<String>()
            : null;

  String get normalizedName => preventKeywords(name);

  String get deprecatedAttribute => deprecated ? '@deprecated' : '';
}

class ListItems implements Typed {
  @override
  final String type;

  @override
  final String ref;

  ListItems.fromJson(Map json)
      : type = json['type'],
        ref = json[r'$ref'];
}

abstract class Typed {
  String get type;
  String get ref;
}

const Set<String> _dartKeywords = {
  'abstract', 'deferred', 'if', 'super',
  'as', 'do', 'implements', 'switch',
  'assert', 'dynamic', 'import', 'sync',
  'async', 'else', 'in', 'this',
  'enum', 'is', 'throw',
  'await', 'export', 'library', 'true',
  'break', 'external', 'new', 'try',
  'case', 'extends', 'null', 'typedef',
  'catch', 'factory', 'operator', 'var',
  'class', 'false', 'part', 'void',
  'const', 'final', 'rethrow', 'while',
  'continue', 'finally', 'return', 'with',
  'covariant', 'for', 'yield',
  'default', 'static' //
};

String preventKeywords(String input) {
  if (_dartKeywords.contains(input)) {
    return '$input\$';
  } else {
    return input;
  }
}
